#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <errno.h>

#include "msr-index.h"

#include <xen/xen.h>
#include <xen/vcpu.h>
#include <xen/event_channel.h>

#include "xen/names.h"
#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

typedef int (*xen_hypercall)(struct xenvcpu *vcpu, uint64_t *args);

/* ------------------------------------------------------------------ */
/* hypercalls: common all                                             */

static int console_io(struct xenvcpu *vcpu, uint64_t *args)
{
    struct xenvm *xen = vcpu->vm;
    char buf[256];
    int count, rc;

    switch (args[0]) {
    case CONSOLEIO_write:
	count = args[1];
	if (count > sizeof(buf))
	    count = sizeof(buf);
	if (0 != (rc = copy_from_guest(vcpu, buf, args[2], count)))
	    return rc;
	conhyp_print(xen, buf, count);
	return count;
    default:
	d0printf("%s: FIXME: unknown %d (%s)\n", __FUNCTION__,
		 (int)args[0], consoleio_name(args[0]));
	return -ENOSYS;
    }
}

static int vcpu_op(struct xenvcpu *vcpu, uint64_t *args)
{
    struct xenvm *xen = vcpu->vm;
    struct xenvcpu *opcpu;
    uint32_t *ptr32;
    uint64_t *ptr64;

    if (args[1] >= xen->vcpus)
	return -ENOENT;
    opcpu = xen->vcpu + args[1];
    
    switch (args[0]) {
    case VCPUOP_register_runstate_memory_area:
	switch (xen->mode) {
	case XENMODE_32:
	case XENMODE_PAE:
	    ptr32 = guest_vaddr_to_ptr(vcpu, args[2]);
	    xen->vcpu[vcpu->id].rs.rs32 = guest_vaddr_to_ptr(vcpu, *ptr32);
	    break;
	case XENMODE_64:
	    ptr64 = guest_vaddr_to_ptr(vcpu, args[2]);
	    xen->vcpu[vcpu->id].rs.rs64 = guest_vaddr_to_ptr(vcpu, *ptr64);
	    break;
	}
	return 0;

    case VCPUOP_is_up:
	return opcpu->online;

    case VCPUOP_set_periodic_timer:
    case VCPUOP_stop_periodic_timer:
    case VCPUOP_set_singleshot_timer:
    case VCPUOP_stop_singleshot_timer:
    case VCPUOP_initialise:
    case VCPUOP_up:
    case VCPUOP_down:
	/* not reached, handled by emu */
	return 0;

    default:
	d0printf("%s/%d: FIXME: unknown %d (%s)\n", __FUNCTION__, vcpu->id,
		 (int)args[0], vcpuop_name(args[0]));
	return -ENOSYS;
    }
    return 0;
}

static int event_channel_op(struct xenvcpu *vcpu, uint64_t *args)
{
    struct xenvm *xen = vcpu->vm;
    int rc;
    
    switch (args[0]) {
#if 0 /* handled by emu now */
    case EVTCHNOP_alloc_unbound:
    {
	struct evtchn_alloc_unbound a;

	if (0 != (rc = copy_from_guest(vcpu, &a, args[1], sizeof(a))))
	    return rc;
	if (a.dom != DOMID_SELF || a.remote_dom != 0)
	    return -EINVAL;
	a.port = evtchn_port(xen, "unbound");
	d1printf("%s: alloc unbound -> port %d\n", __FUNCTION__, a.port);
	if (0 != (rc = copy_to_guest(vcpu, args[1], &a, sizeof(a))))
	    return rc;
	break;
    }
    case EVTCHNOP_bind_virq:
    {
	struct evtchn_bind_virq bind;
	
	if (0 != (rc = copy_from_guest(vcpu, &bind, args[1], sizeof(bind))))
	    return rc;
	if (bind.virq >= NR_VIRQS)
	    return -EINVAL;
	bind.port = evtchn_port(xen, "virq");
	d1printf("%s: bind virq %d (%s) vcpu %d -> port %d\n", __FUNCTION__,
		 bind.virq, GETNAME(virq_names, bind.virq),
		 bind.vcpu, bind.port);
	xen->virq[bind.virq] = bind.port;
	if (0 != (rc = copy_to_guest(vcpu, args[1], &bind, sizeof(bind))))
	    return rc;
	break;
    }
    case EVTCHNOP_bind_ipi:
    {
	struct evtchn_bind_ipi bind;

	if (0 != (rc = copy_from_guest(vcpu, &bind, args[1], sizeof(bind))))
	    return rc;
	bind.port = evtchn_port(xen, "ipi");
	d1printf("%s: bind ipi vcpu %d -> port %d\n", __FUNCTION__,
		 bind.vcpu, bind.port);
	if (0 != (rc = copy_to_guest(vcpu, args[1], &bind, sizeof(bind))))
	    return rc;
	break;
    }
    case EVTCHNOP_unmask:
    {
	struct evtchn_unmask unmask;

	if (0 != (rc = copy_from_guest(vcpu, &unmask, args[1], sizeof(unmask))))
	    return rc;
	evtchn_unmask(xen, unmask.port);
	break;
    }
#endif
    case EVTCHNOP_send:
    {
	struct evtchn_send send;

	if (0 != (rc = copy_from_guest(vcpu, &send, args[1], sizeof(send))))
	    return rc;
	if (0 == evtchn_notify(xen, send.port)) {
	    d2printf("%s: send %d (external, ok)\n", __FUNCTION__, send.port);
	} else {
	    d1printf("%s: send %d (unhandled)\n", __FUNCTION__, send.port);
	}
	break;
    }
    case EVTCHNOP_bind_pirq:
	return -EPERM;
    default:
	d0printf("%s: FIXME: unknown %d (%s)\n", __FUNCTION__,
		 (int)args[0], evtchnop_name(args[0]));
	return -ENOSYS;
    }
    return 0;
}

/* ------------------------------------------------------------------ */
/* hypercall tables                                                   */

static xen_hypercall hcalls_32[XEN_HCALL_MAX] = {
#if 0 /* dead code */
    [ __HYPERVISOR_xen_version ]       = xen_version,
    [ __HYPERVISOR_vm_assist ]         = vm_assist,
    [ __HYPERVISOR_nmi_op ]            = nmi_op,
    [ __HYPERVISOR_physdev_op ]        = physdev_op,
    [ __HYPERVISOR_set_debugreg ]      = set_debugreg,
    [ __HYPERVISOR_fpu_taskswitch ]    = fpu_taskswitch,
    [ __HYPERVISOR_grant_table_op ]    = grant_table_op,
    [ __HYPERVISOR_mmuext_op ]         = mmuext_op_32pae,
    [ __HYPERVISOR_stack_switch ]      = stack_switch_32pae,
    [ __HYPERVISOR_update_descriptor ] = update_descriptor_32pae,
    [ __HYPERVISOR_multicall ]         = multicall_32pae,
    [ __HYPERVISOR_mmu_update ]        = mmu_update_32,
    [ __HYPERVISOR_update_va_mapping ] = update_va_mapping_32,
    [ __HYPERVISOR_sched_op ]          = sched_op,
    [ __HYPERVISOR_memory_op ]         = memory_op,
    [ __HYPERVISOR_set_trap_table ]    = set_trap_table_32pae,
    [ __HYPERVISOR_set_gdt ]           = set_gdt_32pae,
    [ __HYPERVISOR_set_callbacks ]     = set_callbacks_32pae,
    [ __HYPERVISOR_callback_op ]       = callback_op_32pae,
    [ __HYPERVISOR_set_timer_op ]      = set_timer_op_32pae,
#endif

    [ __HYPERVISOR_console_io ]        = console_io,
    [ __HYPERVISOR_vcpu_op ]           = vcpu_op,
    [ __HYPERVISOR_event_channel_op ]  = event_channel_op,
};

static xen_hypercall hcalls_pae[XEN_HCALL_MAX] = {
#if 0 /* dead code */
    [ __HYPERVISOR_xen_version ]       = xen_version,
    [ __HYPERVISOR_vm_assist ]         = vm_assist,
    [ __HYPERVISOR_nmi_op ]            = nmi_op,
    [ __HYPERVISOR_physdev_op ]        = physdev_op,
    [ __HYPERVISOR_set_debugreg ]      = set_debugreg,
    [ __HYPERVISOR_fpu_taskswitch ]    = fpu_taskswitch,
    [ __HYPERVISOR_grant_table_op ]    = grant_table_op,
    [ __HYPERVISOR_mmu_update ]        = mmu_update_pae,
    [ __HYPERVISOR_mmuext_op ]         = mmuext_op_32pae,
    [ __HYPERVISOR_stack_switch ]      = stack_switch_32pae,
    [ __HYPERVISOR_multicall ]         = multicall_32pae,
    [ __HYPERVISOR_update_descriptor ] = update_descriptor_32pae,
    [ __HYPERVISOR_update_va_mapping ] = update_va_mapping_pae,
    [ __HYPERVISOR_sched_op ]          = sched_op,
    [ __HYPERVISOR_memory_op ]         = memory_op,
    [ __HYPERVISOR_set_trap_table ]    = set_trap_table_32pae,
    [ __HYPERVISOR_set_gdt ]           = set_gdt_32pae,
    [ __HYPERVISOR_set_callbacks ]     = set_callbacks_32pae,
    [ __HYPERVISOR_callback_op ]       = callback_op_32pae,
    [ __HYPERVISOR_set_timer_op ]      = set_timer_op_32pae,
#endif

    [ __HYPERVISOR_console_io ]        = console_io,
    [ __HYPERVISOR_vcpu_op ]           = vcpu_op,
    [ __HYPERVISOR_event_channel_op ]  = event_channel_op,
};

#ifdef __x86_64__

static xen_hypercall hcalls_64[XEN_HCALL_MAX] = {
#if 0 /* dead code */
    [ __HYPERVISOR_xen_version ]       = xen_version,
    [ __HYPERVISOR_vm_assist ]         = vm_assist,
    [ __HYPERVISOR_nmi_op ]            = nmi_op,
    [ __HYPERVISOR_physdev_op ]        = physdev_op,
    [ __HYPERVISOR_set_debugreg ]      = set_debugreg,
    [ __HYPERVISOR_fpu_taskswitch ]    = fpu_taskswitch,
    [ __HYPERVISOR_grant_table_op ]    = grant_table_op,
    [ __HYPERVISOR_sched_op ]          = sched_op,
    [ __HYPERVISOR_memory_op ]         = memory_op,
    [ __HYPERVISOR_set_callbacks ]     = set_callbacks_64,
    [ __HYPERVISOR_callback_op ]       = callback_op_64,
    [ __HYPERVISOR_set_trap_table ]    = set_trap_table_64,
    [ __HYPERVISOR_set_segment_base ]  = set_segment_base_64,
    [ __HYPERVISOR_set_gdt ]           = set_gdt_64,
    [ __HYPERVISOR_set_timer_op ]      = set_timer_op_64,
#endif

    [ __HYPERVISOR_console_io ]        = console_io,
    [ __HYPERVISOR_vcpu_op ]           = vcpu_op,
    [ __HYPERVISOR_event_channel_op ]  = event_channel_op,
};

#endif

/* ------------------------------------------------------------------ */

int do_xen_hypercall(struct xenvcpu *vcpu, int nr)
{
    struct xenvm *xen = vcpu->vm;
    uint64_t args[6];
    uint64_t retval = -ENOSYS;
    xen_hypercall hcall;

    need_regs(vcpu);
    need_sregs(vcpu);
    if (nr >= XEN_HCALL_MAX)
	goto done;

    xen->vminfo->hcalls[nr]++;
    d3printf("hypercall #%d (%s)\n", nr, __hypervisor_name(nr));
    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	hcall = (xen->mode == XENMODE_32) ? hcalls_32[nr] : hcalls_pae[nr];
	if (!hcall) {
	    d0printf("unhandled xen hypercall #%d (32/pae, %s)\n",
		     nr, __hypervisor_name(nr));
	    goto done;
	}
	args[0] = vcpu->regs.rbx & 0xffffffff;
	args[1] = vcpu->regs.rcx & 0xffffffff;
	args[2] = vcpu->regs.rdx & 0xffffffff;
	args[3] = vcpu->regs.rsi & 0xffffffff;
	args[4] = vcpu->regs.rdi & 0xffffffff;
	args[5] = vcpu->regs.rbp & 0xffffffff;
	retval = hcall(vcpu, args);
	break;

#ifdef __x86_64__
    case XENMODE_64:
	hcall = hcalls_64[nr];
	if (!hcall) {
	    d0printf("unhandled xen hypercall #%d (64, %s)\n",
		     nr, __hypervisor_name(nr));
	    goto done;
	}
	args[0] = vcpu->regs.rdi;
	args[1] = vcpu->regs.rsi;
	args[2] = vcpu->regs.rdx;
	args[3] = vcpu->regs.r10;
	args[4] = vcpu->regs.r8;
	args[5] = vcpu->regs.r9;
	retval = hcall(vcpu, args);
	break;
#endif
	
    default:
	d0printf("%s: unhandled xen mode %d\n", __FUNCTION__, xen->mode);
	break;
    }
    
done:
    vcpu->regs.rax = retval;
    flush_regs(vcpu);

    return 0;
}

void hypercall_stats(struct xenvm *xen)
{
    print_hcall_stats(xen, xen->vminfo->hcalls, NULL);
    print_fault_stats(xen, xen->vminfo->faults, NULL);
    print_event_stats(xen, xen->vminfo->events, NULL, xen->vminfo->enames);
}
